package Pcore::App::Controller::ExtJS::App;

use Pcore -role, -result;
use Pcore::Util::Text qw[mark_raw];
use Pcore::Util::Data qw[to_json];
use Pcore::ExtJS;
use Pcore::Share::WWW;
use WWW::CDN;
use JavaScript::Packer qw[];

with qw[Pcore::App::Controller::ExtJS];

requires qw[EXT_viewport];

has extjs_ver                   => ( is => 'ro', isa => Str, default => $WWW::CDN::LATEST_VERION->{extjs} );    # ExtJS framewrok version
has extjs_default_theme_classic => ( is => 'ro', isa => Str, default => 'triton' );
has extjs_default_theme_modern  => ( is => 'ro', isa => Str, default => 'triton' );
has extjs_default_locale        => ( is => 'ro', isa => Str, default => 'en' );
has extjs_app_title             => ( is => 'ro', isa => Str, default => sub { ref $_[0] } );

has static_classes => ( is => 'ro', isa => HashRef,   init_arg => undef );
has static_content => ( is => 'ro', isa => ScalarRef, init_arg => undef );

# TODO fix ExtJS overrides version
around run => sub ( $orig, $self, $req ) {

    # if path tail is not empty - fallback to the original method
    if ( $req->{path_tail} ) {
        $self->$orig($req);

        return;
    }

    my $resources = [];

    # FontAwesome
    push $resources->@*, WWW::CDN->fontawesome;

    # ExtJS framework
    my $devel     = $req->{env}->{QUERY_STRING} =~ /\bdevel\b/sm;
    my $ver       = $self->extjs_ver;
    my $framework = 'classic';

    my $extjs_resources;

    # get theme from query
    if ( $req->{env}->{QUERY_STRING} =~ /\btheme=([a-z-]+)/sm ) {
        my $theme = $1;

        $extjs_resources = WWW::CDN->extjs( $ver, $framework, $theme, $devel );
    }

    # fallback to the default theme
    if ( !$extjs_resources ) {
        my $theme = $framework eq 'classic' ? $self->extjs_default_theme_classic : $self->extjs_default_theme_modern;

        $extjs_resources = WWW::CDN->extjs( $ver, $framework, $theme, $devel );
    }

    push $resources->@*, $extjs_resources->@*;

    # ExtJS locale
    my $locale = $self->extjs_default_locale;

    push $resources->@*, WWW::CDN->extjs_locale( $ver, $framework, $locale, $devel );

    # ExtJS overrides
    push $resources->@*, q[<script type="text/javascript" src="/static/extjs/pcore/overrides-v6.js"></script>];

    my $loader_path = {
        q[*]                       => '.',
        Ext                        => '/static/extjs/src/',
        'Ext.ux'                   => '/static/extjs/ux/',
        Pcore                      => '/static/extjs/pcore/',
        $self->{ext_app_namespace} => '.',
    };

    $self->{app}->{api}->{map}->extdirect_map(
        $req->{host},
        undef, undef,
        sub ($api_map) {
            $self->_get_static_content(
                sub ($static_classes) {
                    my $data = {
                        INDEX => {    #
                            title => $self->extjs_app_title
                        },
                        resources => $resources,
                        ext       => {
                            api_map => $api_map ? to_json($api_map)->$* : undef,
                            loader_path    => mark_raw( to_json( $loader_path, readable => $devel )->$* ),
                            app_namespace  => $self->{ext_app_namespace},
                            viewport_class => "$self->{ext_app_namespace}.Viewport",
                            static_classes => $static_classes->$*,
                        },
                    };

                    my $index = P->tmpl->render( 'extjs/index.html', $data );

                    $req->( 200, [ 'Content-Type' => 'text/html; charset=UTF-8' ], $index )->finish;

                    return;
                }
            );

            return;
        }
    );

    return;
};

sub _get_static_content ( $self, $cb ) {
    if ( defined $self->{staic_content} ) {
        $cb->( $self->{staic_content} );

        return;
    }

    my $static_content = [];

    my $js_packer = JavaScript::Packer->init;

    my $cv = AE::cv sub {
        undef $self->{static_classes};

        $self->{staic_content} = \join ';', $static_content->@*, '';

        # compress javascript
        $js_packer->minify( $self->{staic_content}, { compress => 'obfuscate' } );    # clean

        $cb->( $self->{staic_content} );

        return;
    };

    $cv->begin;

    for my $ext_class_name ( keys $self->{static_classes}->%* ) {
        $cv->begin;

        _generate_static_class(
            $self->{static_classes}->{$ext_class_name}->{ctrl},
            $self->{static_classes}->{$ext_class_name}->{class_name},
            $self->{static_classes},
            sub ($js) {
                push $static_content->@*, $js->$*;

                $cv->end;

                return;
            }
        );
    }

    $cv->end;

    return;
}

sub _generate_static_class ( $self, $extjs_class_name, $preloaded_classes, $cb ) {
    my $internal_class_name = lc $extjs_class_name;

    if ( my $extend = $self->{ext_map}->{$internal_class_name}->{extend} ) {
        my $readable = 1;

        # create ExtJS class generation context
        my $ext = bless {
            app        => $self->{app},
            req        => undef,
            readable   => $readable,
            namespace  => $self,
            class_name => $extjs_class_name,
            extend     => $extend,
            cb         => sub ( $ext, $cfg ) {
                $ext->{cfg} = $cfg;

                for my $deps ( keys $ext->{requires}->%* ) {
                    delete $ext->{requires}->{$deps} if exists $preloaded_classes->{$deps};
                }

                # generate class
                $cb->( $ext->to_js );

                return;
            },
          },
          'Pcore::ExtJS';

        my $method = "EXT_$internal_class_name";

        # call ExtJS class generation method
        $self->$method($ext);

        return;
    }

    return;
}

1;
## -----SOURCE FILTER LOG BEGIN-----
##
## PerlCritic profile "pcore-script" policy violations:
## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
## | Sev. | Lines                | Policy                                                                                                         |
## |======+======================+================================================================================================================|
## |    3 | 161                  | Subroutines::ProhibitManyArgs - Too many arguments                                                             |
## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
## |    2 | 127                  | ValuesAndExpressions::ProhibitEmptyQuotes - Quotes used with a string containing no non-whitespace characters  |
## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
## |    1 | 47                   | RegularExpressions::ProhibitEnumeratedClasses - Use named character classes ([a-z] vs. [[:lower:]])            |
## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
##
## -----SOURCE FILTER LOG END-----
__END__
=pod

=encoding utf8

=head1 NAME

Pcore::App::Controller::ExtJS::App - ExtJS application HTTP controller

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 ATTRIBUTES

=head1 METHODS

=head1 SEE ALSO

=cut
